/*
 * jSite2 - FcpCollector.java -
 * Copyright © 2008 David Roden
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package net.pterodactylus.jsite.core;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.pterodactylus.fcp.highlevel.HighLevelClient;
import net.pterodactylus.fcp.highlevel.HighLevelClientListener;
import net.pterodactylus.fcp.highlevel.HighLevelException;
import net.pterodactylus.fcp.highlevel.KeyGenerationResult;
import net.pterodactylus.jsite.util.IdGenerator;
import net.pterodactylus.util.io.Closer;
import net.pterodactylus.util.logging.Logging;
import net.pterodactylus.util.number.Hex;

/**
 * TODO
 * 
 * @author David ‘Bombe’ Roden &lt;bombe@freenetproject.org&gt;
 */
public class NodeManager implements Iterable<Node>, PropertyChangeListener, HighLevelClientListener {

	/** Logger. */
	private static final Logger logger = Logging.getLogger(NodeManager.class.getName());

	/** The FCP client name. */
	private final String clientName;

	/** The directory for the configuration. */
	private final String directory;

	/** Object used for synchronization. */
	private final Object syncObject = new Object();

	/** Node listeners. */
	private List<NodeListener> nodeListeners = Collections.synchronizedList(new ArrayList<NodeListener>());

	/** All nodes. */
	private List<Node> nodes = Collections.synchronizedList(new ArrayList<Node>());

	/** Map from node ID to node. */
	private Map<String, Node> idNodes = Collections.synchronizedMap(new HashMap<String, Node>());

	/** All FCP connections. */
	private Map<Node, HighLevelClient> nodeClients = Collections.synchronizedMap(new HashMap<Node, HighLevelClient>());

	/** Maps nodes to high-level clients. */
	private Map<HighLevelClient, Node> clientNodes = Collections.synchronizedMap(new HashMap<HighLevelClient, Node>());

	/**
	 * Creates a new FCP collector.
	 * 
	 * @param clientName
	 *            The name of the FCP client
	 * @param directory
	 *            The directory in which to store the nodes
	 */
	public NodeManager(String clientName, String directory) {
		this.clientName = clientName;
		this.directory = directory;
	}

	//
	// EVENT MANAGEMENT
	//

	/**
	 * Adds the given listener to the list of listeners.
	 * 
	 * @param nodeListener
	 *            The listener to add
	 */
	public void addNodeListener(NodeListener nodeListener) {
		nodeListeners.add(nodeListener);
	}

	/**
	 * Removes the given listener from the list of listeners.
	 * 
	 * @param nodeListener
	 *            The listener to remove
	 */
	public void removeNodeListener(NodeListener nodeListener) {
		nodeListeners.remove(nodeListener);
	}

	/**
	 * Notifies all listeners that a node was added.
	 * 
	 * @param node
	 *            The node that was added.
	 */
	private void fireNodeAdded(Node node) {
		for (NodeListener nodeListener : nodeListeners) {
			nodeListener.nodeAdded(node);
		}
	}

	/**
	 * Notifies all listeners that a node was removed.
	 * 
	 * @param node
	 *            The node that was removed
	 */
	private void fireNodeRemoved(Node node) {
		for (NodeListener nodeListener : nodeListeners) {
			nodeListener.nodeRemoved(node);
		}
	}

	/**
	 * Notifies all listeners that the given node was connected.
	 * 
	 * @param node
	 *            The node that is now connected
	 */
	private void fireNodeConnected(Node node) {
		for (NodeListener nodeListener : nodeListeners) {
			nodeListener.nodeConnected(node);
		}
	}

	/**
	 * Notifies all listeners that a connection to a node has failed.
	 * 
	 * @param node
	 *            The node that could not be connected
	 * @param cause
	 *            The cause of the failure
	 */
	private void fireNodeConnectionFailed(Node node, Throwable cause) {
		for (NodeListener nodeListener : nodeListeners) {
			nodeListener.nodeConnectionFailed(node, cause);
		}
	}

	/**
	 * Notifies all listeners that the given node was disconnected.
	 * 
	 * @param node
	 *            The node that is now disconnected
	 * @param throwable
	 *            The exception that caused the disconnect, or <code>null</code>
	 *            if there was no exception
	 */
	private void fireNodeDisconnected(Node node, Throwable throwable) {
		for (NodeListener nodeListener : nodeListeners) {
			nodeListener.nodeDisconnected(node, throwable);
		}
	}

	//
	// ACCESSORS
	//

	/**
	 * Returns the directory in which the nodes are stored.
	 * 
	 * @return The directory the nodes are stored in
	 */
	public String getDirectory() {
		return directory;
	}

	/**
	 * Checks whether the given node is already connected.
	 * 
	 * @param node
	 *            The node to check
	 * @return <code>true</code> if the node is already connected,
	 *         <code>false</code> otherwise
	 */
	public boolean hasNode(Node node) {
		return nodes.contains(node);
	}

	/**
	 * {@inheritDoc}
	 */
	public Iterator<Node> iterator() {
		return nodes.iterator();
	}

	//
	// ACTIONS
	//

	/**
	 * Loads nodes.
	 * 
	 * @throws IOException
	 *             if an I/O error occurs loading the nodes
	 */
	public void load() throws IOException {
		File directoryFile = new File(directory);
		File nodeFile = new File(directoryFile, "nodes.properties");
		if (!nodeFile.exists() || !nodeFile.isFile() || !nodeFile.canRead()) {
			return;
		}
		Properties nodeProperties = new Properties();
		InputStream nodeInputStream = null;
		try {
			nodeInputStream = new FileInputStream(nodeFile);
			nodeProperties.load(nodeInputStream);
		} finally {
			Closer.close(nodeInputStream);
		}
		int nodeIndex = -1;
		List<Node> loadedNodes = new ArrayList<Node>();
		while (nodeProperties.containsKey("nodes." + ++nodeIndex + ".name")) {
			String nodePrefix = "nodes." + nodeIndex;
			String nodeId = nodeProperties.getProperty(nodePrefix + ".id");
			if (nodeId == null) {
				nodeId = Hex.toHex(IdGenerator.generateId());
			}
			String nodeName = nodeProperties.getProperty(nodePrefix + ".name");
			if (!Verifier.verifyNodeName(nodeName)) {
				logger.log(Level.WARNING, "invalid node name “" + nodeName + "”, skipping…");
				continue;
			}
			String nodeHostname = nodeProperties.getProperty(nodePrefix + ".hostname");
			if (!Verifier.verifyHostname(nodeHostname)) {
				logger.log(Level.WARNING, "invalid host name “" + nodeHostname + "”");
				/* not fatal, might be valid later on. */
			}
			String nodePortString = nodeProperties.getProperty(nodePrefix + ".port");
			if (!Verifier.verifyPort(nodePortString)) {
				logger.log(Level.WARNING, "invalid port number “" + nodePortString + "”, skipping…");
				continue;
			}
			int nodePort = -1;
			try {
				nodePort = Integer.valueOf(nodePortString);
			} catch (NumberFormatException nfe1) {
				/* shouldn't happen, port number was checked before. */
				logger.log(Level.SEVERE, "invalid port number “" + nodePortString + "”, check failed! skipping…");
				continue;
			}
			Node newNode = new Node();
			newNode.setId(nodeId);
			newNode.setName(nodeName);
			newNode.setHostname(nodeHostname);
			newNode.setPort(nodePort);
			loadedNodes.add(newNode);
		}
		logger.fine("loaded " + loadedNodes.size() + " nodes from config");
		synchronized (syncObject) {
			nodes.clear();
			for (Node node : loadedNodes) {
				addNode(node);
			}
		}
	}

	/**
	 * Saves all configured nodes.
	 * 
	 * @throws IOException
	 *             if an I/O error occurs saving the nodes
	 */
	public void save() throws IOException {
		File directoryFile = new File(directory);
		if (!directoryFile.exists()) {
			if (!directoryFile.mkdirs()) {
				throw new IOException("could not create directory: " + directory);
			}
		}
		Properties nodeProperties = new Properties();
		int nodeIndex = -1;
		for (Node node : nodes) {
			String nodePrefix = "nodes." + ++nodeIndex;
			nodeProperties.setProperty(nodePrefix + ".id", node.getId());
			nodeProperties.setProperty(nodePrefix + ".name", node.getName());
			nodeProperties.setProperty(nodePrefix + ".hostname", node.getHostname());
			nodeProperties.setProperty(nodePrefix + ".port", String.valueOf(node.getPort()));
		}
		File projectFile = new File(directoryFile, "nodes.properties");
		OutputStream nodeOutputStream = null;
		try {
			nodeOutputStream = new FileOutputStream(projectFile);
			nodeProperties.store(nodeOutputStream, "jSite nodes");
		} finally {
			Closer.close(nodeOutputStream);
		}
	}

	/**
	 * Adds the given node to this manager.
	 * 
	 * @see #connect(Node)
	 * @param node
	 *            The node to connect to
	 * @return <code>true</code> if the node was added, <code>false</code>
	 *         if the node was not added because it was already known
	 */
	public boolean addNode(Node node) {
		if (nodes.contains(node)) {
			logger.warning("was told to add already known node: " + node);
			return false;
		}
		node.addPropertyChangeListener(this);
		HighLevelClient highLevelClient = new HighLevelClient(clientName);
		nodes.add(node);
		idNodes.put(node.getId(), node);
		clientNodes.put(highLevelClient, node);
		nodeClients.put(node, highLevelClient);
		highLevelClient.addHighLevelClientListener(this);
		fireNodeAdded(node);
		return true;
	}

	/**
	 * Removes the given node from the node manager, disconnecting it if it is
	 * currently connected.
	 * 
	 * @param node
	 *            The node to remove
	 */
	public void removeNode(Node node) {
		synchronized (syncObject) {
			if (!nodes.contains(node)) {
				return;
			}
			if (nodeClients.containsKey(node)) {
				disconnect(node);
			}
			nodes.remove(node);
			idNodes.remove(node.getId());
			node.removePropertyChangeListener(this);
			fireNodeRemoved(node);
		}
	}

	/**
	 * Tries to establish a connection with the given node.
	 * 
	 * @param node
	 *            The node to connect to
	 */
	public void connect(Node node) {
		HighLevelClient highLevelClient;
		highLevelClient = nodeClients.get(node);
		if (highLevelClient == null) {
			logger.warning("was told to connect to unknown node: " + node);
			return;
		}
		try {
			highLevelClient.connect(node.getHostname(), node.getPort());
		} catch (UnknownHostException uhe1) {
			fireNodeConnectionFailed(node, uhe1);
		} catch (IOException ioe1) {
			fireNodeConnectionFailed(node, ioe1);
		}
	}

	/**
	 * Disconnects the given node without removing it.
	 * 
	 * @param node
	 *            The node to disconnect
	 */
	public void disconnect(Node node) {
		synchronized (syncObject) {
			if (!nodes.contains(node)) {
				return;
			}
			HighLevelClient highLevelClient = nodeClients.get(node);
			highLevelClient.disconnect();
		}
	}

	/**
	 * Returns a list of all nodes.
	 * 
	 * @return A list of all nodes
	 */
	public List<Node> getNodes() {
		return Collections.unmodifiableList(nodes);
	}

	/**
	 * Returns the high-level client for a given node.
	 * 
	 * @param node
	 *            The node to get a high-level client for
	 * @return The high-level client for a node, or <code>null</code> if the
	 *         node was disconnected or removed
	 */
	public HighLevelClient getHighLevelClient(Node node) {
		return nodeClients.get(node);
	}

	/**
	 * Returns the node for a high-level client.
	 * 
	 * @param highLevelClient
	 *            The high-level client to get the node for
	 * @return The node for the high-level client, or <code>null</code> if the
	 *         high-level client is not known
	 */
	public Node getNode(HighLevelClient highLevelClient) {
		return clientNodes.get(highLevelClient);
	}

	/**
	 * Returns the node identified by the given ID.
	 * 
	 * @param id
	 *            The ID of the node
	 * @return The node with the given ID, or <code>null</code> if no such
	 *         node was found
	 */
	Node getNode(String id) {
		return idNodes.get(id);
	}

	/**
	 * Generates a new SSK key pair.
	 * 
	 * @return An array with the private key at index <code>0</code> and the
	 *         public key at index <code>1</code>
	 * @throws IOException
	 *             if an I/O error occurs communicating with the node
	 * @throws JSiteException
	 *             if there is a problem with the node
	 */
	public String[] generateKeyPair() throws IOException, JSiteException {
		if (nodes.isEmpty()) {
			throw new NoNodeException("no node configured");
		}
		Node node = nodes.get(0);
		HighLevelClient highLevelClient = nodeClients.get(node);
		try {
			KeyGenerationResult keyGenerationResult = highLevelClient.generateKey().getResult();
			return new String[] { keyGenerationResult.getInsertURI(), keyGenerationResult.getRequestURI() };
		} catch (HighLevelException hle1) {
			throw new BackendException(hle1);
		} catch (InterruptedException e) {
			/* ignore. */
		}
		return null;
	}

	//
	// PRIVATE METHODS
	//

	//
	// INTERFACE HighLevelClientListener
	//

	/**
	 * {@inheritDoc}
	 */
	public void clientConnected(HighLevelClient highLevelClient) {
		logger.log(Level.FINER, "clientConnected(c=" + highLevelClient + ")");
		Node node = clientNodes.get(highLevelClient);
		if (node == null) {
			logger.log(Level.WARNING, "got event for unknown client");
			return;
		}
		fireNodeConnected(node);
	}

	/**
	 * {@inheritDoc}
	 */
	public void clientDisconnected(HighLevelClient highLevelClient, Throwable throwable) {
		logger.log(Level.FINER, "clientDisconnected(c=" + highLevelClient + ",t=" + throwable + ")");
		synchronized (syncObject) {
			Node node = clientNodes.get(highLevelClient);
			if (node == null) {
				logger.log(Level.WARNING, "got event for unknown client");
				return;
			}
			fireNodeDisconnected(node, throwable);
		}
	}

	//
	// INTERFACE PropertyChangeListener
	//

	/**
	 * {@inheritDoc}
	 */
	public void propertyChange(PropertyChangeEvent propertyChangeEvent) {
		Object eventSource = propertyChangeEvent.getSource();
		if (eventSource instanceof Node) {
			String propertyName = propertyChangeEvent.getPropertyName();
			if ("hostname".equals(propertyName) || "port".equals(propertyName)) {
				Node node = (Node) eventSource;
				HighLevelClient highLevelClient = nodeClients.get(node);
				if (highLevelClient == null) {
					logger.log(Level.WARNING, "got property change event for unknown node: " + node);
					return;
				}
				if (highLevelClient.isConnected()) {
					highLevelClient.disconnect();
					try {
						highLevelClient.connect(node.getHostname(), node.getPort());
					} catch (UnknownHostException uhe1) {
						fireNodeConnectionFailed(node, uhe1);
					} catch (IOException ioe1) {
						fireNodeConnectionFailed(node, ioe1);
					}
				}
			}
		}
	}

}
